"""
Tests for generate.py
"""

import unittest

from mazenv.generate import prim

class GeneratorTest(unittest.TestCase):
    """
    Tests for maze generators.
    """
    # pylint: disable=R0914
    def test_prim(self):
        """
        Test the prim generator.
        """
        self._test_generators([
            lambda: prim((10, 10)),
            lambda: prim((8, 8, 8)),
            lambda: prim((3, 5, 8, 15))
        ])

    def _test_generators(self, gen_funcs):
        """
        Test desirable properties of maze generators.
        """
        for func in gen_funcs:
            seen = []
            for _ in range(10):
                maze = func()
                self.assertTrue(maze.solve())
                self.assertTrue(maze not in seen)
                seen.append(maze)

if __name__ == '__main__':
    unittest.main()



"""
RL environments for mazes.
"""

import itertools

import gym
import gym.spaces as spaces
import numpy as np

from . import _util as util

NUM_CELL_FIELDS = 5
SPACE_CELL_FIELD = 0
WALL_CELL_FIELD = 1
START_CELL_FIELD = 2
END_CELL_FIELD = 3
CURRENT_CELL_FIELD = 4
ACTION_NOP = 0

class Env(gym.Env):
    """
    Base class for maze environments.

    Observations are the entire maze, represented as a
    Tensor with different entries corresponding to
    different types of objects.

    Actions are discrete with 2*num_dims + 1 options.
    This covers every dimension and a NOP.
    """
    def __init__(self, maze):
        assert maze.start_pos
        obs_shape = maze.shape + (NUM_CELL_FIELDS,)
        self.maze = maze
        self.observation_space = spaces.Box(0, 1, shape=obs_shape)
        self.action_space = spaces.Discrete(len(maze.shape)*2 + 1)
        self.position = maze.start_pos

    def _reset(self):
        self.position = self.maze.start_pos
        return self._make_observation()

    def _step(self, action):
        if action != ACTION_NOP:
            new_pos = list(util.iterate_neighbors(self.position))[action-1]
            if not self.maze.is_wall(new_pos):
                self.position = new_pos
        done = (self.position == self.maze.end_pos)
        rew = -1.0
        if done:
            rew = 0.0
        return self._make_observation(), rew, done, {}

    def _make_observation(self):
        """
        Create an observation for the current state.
        """
        obs = np.zeros(self.observation_space.low.shape, dtype='uint8')
        for position in self.maze.positions():
            self._fill_cell(obs[position], position)
        return obs

    def _fill_cell(self, cell, cell_position):
        """
        Set the entries of the observation cell.
        """
        if self.maze.is_wall(cell_position):
            cell[WALL_CELL_FIELD] = 1
        elif cell_position == self.maze.start_pos:
            cell[START_CELL_FIELD] = 1
        elif cell_position == self.maze.end_pos:
            cell[END_CELL_FIELD] = 1
        else:
            cell[SPACE_CELL_FIELD] = 1
        if cell_position == self.position:
            cell[CURRENT_CELL_FIELD] = 1

class HorizonEnv(gym.ObservationWrapper):
    """
    Wrap the observations of an Env so that they extend a
    fixed distance in every direction.

    The number of cells away you can see along an axis is
    called the "horizon".
    For a horizon of 1, observations have side length 3.
    """
    def __init__(self, env, horizon=1):
        super(HorizonEnv, self).__init__(env)
        self.horizon = horizon
        self.old_shape = env.observation_space.low.shape[:-1]
        num_dims = len(self.old_shape)
        obs_size = (horizon*2 + 1,) * num_dims + (NUM_CELL_FIELDS,)
        self.observation_space = spaces.Box(0, 1, shape=obs_size)

    def _observation(self, observation):
        wall = [0] * NUM_CELL_FIELDS
        wall[WALL_CELL_FIELD] = 1
        center = _obs_current_position(observation)
        res = []
        for position in self._position_indices(center):
            if util.shape_contains(self.old_shape, position):
                res.append(observation[position])
            else:
                res.append(wall)
        return np.array(res).reshape(self.observation_space.low.shape)

    def _position_indices(self, center_pos):
        """
        Compute the indices of each cell in the visible
        horizon grid.

        The cells are generated in order such that they
        can be reshaped to the N-d observation.
        """
        ranges = [list(range(x-self.horizon, x+self.horizon+1)) for x in center_pos]
        return itertools.product(*ranges)

def _obs_current_position(observation):
    """
    Extract the current position from an observation.
    """
    where = np.where(observation[..., CURRENT_CELL_FIELD] == 1)
    return tuple(x[0] for x in where)



"""
Tests for env.py
"""

import unittest

from mazenv.env import CURRENT_CELL_FIELD, Env, HorizonEnv
from mazenv.maze import parse_2d_maze

class EnvTest(unittest.TestCase):
    """
    Tests for raw environments.
    """
    # pylint: disable=R0914
    def test_2d_trajectory(self):
        """
        Test that all the positions, rewards, and done
        signals are correct throughout a trajectory.
        """
        maze = parse_2d_maze('w...w\n' +
                             '..w.w\n' +
                             'Awx..\n' +
                             '..w..')
        env = Env(maze)
        env.reset()
        act_nop, act_up, act_down, act_left, act_right = [0, 1, 2, 3, 4]
        actions = [act_down, act_right, act_right, act_nop, act_left,
                   act_left, act_up, act_up, act_up, act_right, act_up,
                   act_right, act_right, act_down, act_left, act_down,
                   act_nop, act_left]
        positions = [(3, 0), (3, 1), (3, 1), (3, 1), (3, 0),
                     (3, 0), (2, 0), (1, 0), (1, 0), (1, 1), (0, 1),
                     (0, 2), (0, 3), (1, 3), (1, 3), (2, 3),
                     (2, 3), (2, 2)]
        dones = [False]*(len(positions)-1) + [True]
        for act, expected_pos, expected_done in zip(actions, positions, dones):
            obs, rew, done, _ = env.step(act)
            self.assertEqual(done, expected_done)
            if done:
                self.assertEqual(rew, 0.0)
            else:
                self.assertEqual(rew, -1.0)
            self._assert_unique_cell(maze, obs, expected_pos, CURRENT_CELL_FIELD)

    def _assert_unique_cell(self, maze, obs, pos, cell_idx):
        """
        Assert that the cell is set at the position in the
        observation, and nowhere else in the observation.
        """
        for some_pos in maze.positions():
            cell_val = obs[some_pos][cell_idx]
            if some_pos == pos:
                self.assertEqual(cell_val, 1)
            else:
                self.assertEqual(cell_val, 0)

class HorizonEnvTest(unittest.TestCase):
    """
    Tests for limited-horizon environments.
    """
    def test_2d_observations(self):
        """
        Test observations on a 2-D environment.
        """
        maze = parse_2d_maze('w...w\n' +
                             '..w.w\n' +
                             'Awx..\n' +
                             '.....')
        env = HorizonEnv(Env(maze), horizon=1)
        env.reset()
        act_up, act_down, act_left, act_right = [1, 2, 3, 4]
        actions = [act_left, act_down, act_right, act_right, act_up]
        obses = [
            _centered_horizon_obs('w..\nwAw\nw..'),
            _centered_horizon_obs('wAw\nw..\nwww'),
            _centered_horizon_obs('Awx\n...\nwww'),
            _centered_horizon_obs('wx.\n...\nwww'),
            _centered_horizon_obs('.w.\nwx.\n...')
        ]
        for action, expected_obs in zip(actions, obses):
            obs, _, _, _ = env.step(action)
            self.assertTrue((obs == expected_obs).all())

def _centered_horizon_obs(maze_str):
    maze = parse_2d_maze(maze_str)

    center = tuple(x//2 for x in maze.shape)
    old_start = maze.start_pos
    maze.start_pos = center
    env = Env(maze)
    env.reset()
    maze.start_pos = old_start

    obs, _, _, _ = env.step(0)
    return obs

if __name__ == '__main__':
    unittest.main()



"""
Routines for generating random mazes.
"""

import random

import numpy as np

from .maze import Maze
from . import _util as util

def prim(shape):
    """
    Use randomized Prim's algorithm to generate a maze.
    """
    start_pos = tuple([random.randrange(x) for x in shape])
    walls = np.ones(shape, dtype='bool')
    walls[start_pos] = False
    maze = Maze(walls, start_pos=start_pos)

    edges = [x for x in util.iterate_neighbors(start_pos) if maze.in_bounds(x)]
    visited = set(edges)
    visited.add(start_pos)

    while edges:
        pos = random.choice(edges)
        edges.remove(pos)
        if len(list(maze.neighboring_non_walls(pos))) > 1:
            continue
        maze.walls[pos] = False
        for neighbor in util.iterate_neighbors(pos):
            if maze.in_bounds(neighbor) and neighbor not in visited:
                visited.add(neighbor)
                edges.append(neighbor)

    maze.end_pos = random.choice(list(maze.spaces()))
    return maze



"""
APIs for using and manipulating mazes.
"""

from collections import deque

import numpy as np

from . import _util as util

class Maze:
    """
    A rectangular maze.

    Mazes are grids of cell.
    Each cell is either a wall or a space.
    Two spaces, as designated by the start_pos and end_pos
    attributes, represent the start and end cells.
    """
    def __init__(self, wall_array, start_pos=None, end_pos=None):
        """
        Create a maze from the given specifications.

        The wall_array should be a boolean numpy array
        indicating which cells are walls.

        The start_pos and end_pos, if not None, should be
        tuples specifying the coordintaes of the start and
        end spaces.
        """
        if start_pos:
            assert not wall_array[start_pos]
        if end_pos:
            assert not wall_array[end_pos]
        self.walls = wall_array
        self.start_pos = start_pos
        self.end_pos = end_pos

    @property
    def shape(self):
        """
        Get the shape of the maze as a tuple.
        """
        return self.walls.shape

    def in_bounds(self, pos):
        """
        Check that a coordinate tuple is in bounds.
        """
        return util.shape_contains(self.shape, pos)

    def positions(self):
        """
        Return an iterable of all the valid positions on
        the maze, in a deterministic order.
        """
        return util.iterate_positions(self.shape)

    def is_wall(self, pos):
        """
        Check if the position is a wall.

        Always returns True for out-of-bounds positions.
        """
        if self.in_bounds(pos):
            return self.walls[pos]
        return True

    def solve(self):
        """
        Compute an optimal path from the start position to
        the end position.

        Returns a list of tuples, one for each position.
        This list includes the start and end positions.

        If no solution exists, None is returned.
        """
        assert self.start_pos
        if not self.end_pos:
            return None
        path_queue = deque([[(self.start_pos)]])
        visited = set()
        while path_queue:
            path = path_queue.popleft()
            for neighbor in self.neighboring_non_walls(path[-1]):
                if neighbor in visited:
                    continue
                visited.add(neighbor)
                new_path = path.copy()
                new_path.append(neighbor)
                if neighbor == self.end_pos:
                    return new_path
                path_queue.append(new_path)
        return None

    def neighboring_non_walls(self, pos):
        """
        Find the non-walls neighboring the given position.
        """
        for neighbor in util.iterate_neighbors(pos):
            if not self.is_wall(neighbor):
                yield neighbor

    def spaces(self):
        """
        Iterate through all the spaces with no start or
        end positions.
        """
        for pos in self.positions():
            if not (self.is_wall(pos) or
                    pos == self.start_pos or
                    pos == self.end_pos):
                yield pos

    def __str__(self):
        if len(self.walls.shape) != 2:
            return str(self.walls)
        row_strs = []
        for row in range(self.walls.shape[0]):
            row_str = ''
            for col in range(self.walls.shape[1]):
                pos = (row, col)
                if self.is_wall(pos):
                    row_str += 'w'
                elif self.start_pos == pos:
                    row_str += 'A'
                elif self.end_pos == pos:
                    row_str += 'x'
                else:
                    row_str += '.'
            row_strs.append(row_str)
        return '\n'.join(row_strs)

    def __eq__(self, other):
        return (self.start_pos == other.start_pos and
                self.end_pos == other.end_pos and
                self.shape == other.shape and
                (self.walls == other.walls).all())

def parse_2d_maze(maze_str):
    """
    Decode a 2-D maze from its string representation.
    """
    lines = [x.strip() for x in maze_str.strip().split('\n')]
    num_rows = len(lines)
    if num_rows == 0:
        raise ValueError('must have at least one row')
    num_cols = len(lines[0])
    walls = []
    start_pos = None
    end_pos = None
    for row, row_str in enumerate(lines):
        if len(row_str) != num_cols:
            raise ValueError('row length should be %d but got %d' %
                             (num_cols, len(row)))
        sub_walls = []
        for col, cell_str in enumerate(row_str):
            if cell_str == 'w':
                sub_walls.append(True)
            else:
                sub_walls.append(False)
            if cell_str == 'A':
                if start_pos:
                    raise ValueError('more than one start state')
                start_pos = (row, col)
            elif cell_str == 'x':
                if end_pos:
                    raise ValueError('more than one end state')
                end_pos = (row, col)
        walls.append(sub_walls)
    return Maze(np.array(walls), start_pos=start_pos, end_pos=end_pos)



"""
Tests for maze.py
"""

import unittest
import numpy as np

from mazenv.maze import Maze, parse_2d_maze

class TestMazeSerialization(unittest.TestCase):
    """
    Test cases for conversions between mazes and strings.
    """
    def test_str_2d(self):
        """
        Test 2-D maze serialization.
        """
        for maze, expected in _testing_maze_pairs():
            self.assertEqual(str(maze), expected)

    def test_parse(self):
        """
        Test 2-D maze parsing.
        """
        for maze, in_str in _testing_maze_pairs():
            parsed = parse_2d_maze(in_str)
            self.assertEqual(parsed.start_pos, maze.start_pos)
            self.assertEqual(parsed.end_pos, maze.end_pos)
            self.assertTrue((parsed.walls == maze.walls).all())

class TestMazeSolver(unittest.TestCase):
    """
    Test cases for maze solving.
    """
    def test_solvable_2d(self):
        """
        Test on solvable 2-D mazes.
        """
        mazes = [
            parse_2d_maze('A.www\n' + 'wxwww'),
            parse_2d_maze('.w....\n' + 'w..wxw\n' + '.Awwww')
        ]
        solutions = [
            [(0, 0), (0, 1), (1, 1)],
            [(2, 1), (1, 1), (1, 2), (0, 2), (0, 3), (0, 4), (1, 4)]
        ]
        for maze, solution in zip(mazes, solutions):
            self.assertEqual(maze.solve(), solution)

    def test_unsolvable_2d(self):
        """
        Test on unsolvable 2-D mazes.
        """
        mazes = [
            parse_2d_maze('Awwww\n' + 'wxwww'),
            parse_2d_maze('A.www\n' + 'wwwww'),
            parse_2d_maze('.w..w.\n' + 'w..wxw\n' + '.Awwww'),
        ]
        for maze in mazes:
            self.assertEqual(maze.solve(), None)

def _testing_maze_pairs():
    """
    Return tuples of (maze, str).
    """
    res = []
    maze = Maze(np.array([[False, True, False], [True, False, False]]),
                start_pos=(0, 2))
    res.append((maze, '.wA\nw..'))
    maze = Maze(np.array([[False, True], [True, False], [False, True]]),
                start_pos=(1, 1), end_pos=(2, 0))
    res.append((maze, '.w\nwA\nxw'))
    maze = Maze(np.array([[False, True], [True, False], [False, True]]),
                end_pos=(2, 0))
    res.append((maze, '.w\nw.\nxw'))
    return res

if __name__ == '__main__':
    unittest.main()



"""
Utilities for grids.
"""

import itertools

def iterate_positions(shape):
    """
    Generate an iterable of all valid indices in a tensor.
    """
    return itertools.product(*[range(n) for n in shape])

def iterate_neighbors(pos):
    """
    Iterate through all the single-axis neighbors of a
    coordinate tuple.

    Order is deterministic.
    """
    for axis, val in enumerate(pos):
        for offset in [-1, 1]:
            yield pos[:axis] + (val+offset,) + pos[axis+1:]

def shape_contains(shape, pos):
    """
    Check if a tensor of the given shape contains the
    position.
    """
    assert len(shape) == len(pos)
    for val, max_val in zip(pos, shape):
        if val < 0 or val >= max_val:
            return False
    return True
